-- Cached function references (for performance).
local byte   = string.byte
local find   = string.find
local match  = string.match
local gsub   = string.gsub
local format = string.format
local insert = table.insert
local sort   = table.sort
local concat = table.concat
-- For escaping string values.
local str_escape_map = {
    ['\a'] = '\\a', ['\b'] = '\\b', ['\t'] = '\\t', ['\n'] = '\\n',
    ['\v'] = '\\v', ['\f'] = '\\f', ['\r'] = '\\r', ['\\'] = '\\\\' }
local str_escape_replace = function(c)
    return str_escape_map[c] or format('\\%03d', byte(c))
end
-- Keys are comparable only if the same type, otherwise just sort them by type.
local types_order, ref_types_order = {
    ['number'] = 0, ['boolean'] = 1, ['string'] = 2, ['table'] = 3,
    ['function'] = 4 }, 5
local function compare_keys(k1, k2)
    local t1, t2 = type(k1), type(k2)
    if t1 ~= t2 then -- not the same type
        return (types_order[t1] or ref_types_order)
             < (types_order[t2] or ref_types_order)
    elseif t1 == 'boolean' then -- comparing booleans
        return not k1 -- Sort false before true.
    elseif t1 == 'number' or t1 == 'string' then -- comparing numbers (including NaNs or infinites) or strings
        return k1 < k2 -- Keys with the same comparable type.
    else -- comparing references (including tables, functions, userdata, threads...)
        return tostring(k1) < tostring(k2) -- may be the Lua engine adds some comparable info
    end
end
-- String keys matching valid identifiers that are reserved by Lua.
local reserved_keys = {
    ['and']      = 1, ['break'] = 1, ['do']    = 1, ['else']   = 1,
    ['elseif']   = 1, ['end']   = 1, ['false'] = 1, ['for']    = 1,
    ['function'] = 1, ['if']    = 1, ['in']    = 1, ['local']  = 1,
    ['nil']      = 1, ['not']   = 1, ['or']    = 1, ['repeat'] = 1,
    ['return']   = 1, ['then']  = 1, ['true']  = 1, ['until']  = 1,
    ['while']    = 1 }
-- Main function.
val_to_str = function(val, options)
    -- Decode and cache the options.
    local include_mt  = options and options.include_mt
    local prettyprint = options and options.prettyprint
    local asciionly   = options and options.asciionly
    -- Precompute the output formats depending on options.
    local open   = prettyprint and '{ '  or '{'
    local equals = prettyprint and ' = ' or '='
    local comma  = prettyprint and ', '  or ','
    local close  = prettyprint and ' }'  or '}'
    -- What to escape: C0 controls, the backslash, and optionally non-ASCII bytes.
    local str_escape_pattern = asciionly and '[%z\001-\031\\\127-\255]' or '[%z\001-\031\\\127]'
     -- Indexed references (mapped to ids), and counters per ref type.
    local ref_ids, ref_counts = {}, {}
    -- Helper needed to detect recursive tables and avoid infinite loops.
    local function visit(ref)
        local typ = type(ref)
        if typ == 'number' or typ == 'boolean' then
            return tostring(ref)
        elseif typ == 'string' then
                local escaped = format("%q", ref);
                return(escaped)
--[[           if find(ref, "'") then
               str_escape_map['"'] = '\\"'
               return '"' .. gsub(ref, str_escape_pattern, str_escape_replace) .. '"'
            else
               str_escape_map['"'] = '"'
               return "'" .. gsub(ref, str_escape_pattern, str_escape_replace) .. "'"
            end
]]--
        elseif typ == 'table' then
	        local id = ref_ids[ref]
            if id then
                return ':' .. typ .. '#' .. id .. ':'
            end
            id = (ref_counts[typ] or 0) + 1; ref_ids[ref], ref_counts[typ] = id, id
            -- First dump keys that are in sequence.
            local result, sequenced, keys = {}, {}, {}
            for i, val in ipairs(ref) do
                insert(result, visit(val))
                sequenced[i] = true
            end
            -- Then dump other keys out of sequence, in a stable order.
            for key, _ in pairs(ref) do
                if not sequenced[key] then
                    insert(keys, key)
                end
            end
            sequenced = nil -- Free the temp table no longer needed.
            -- Sorting keys (of any type) is needed for stable comparison of results.
            sort(keys, compare_keys)
            for _, key in ipairs(keys) do
                insert(result,
                    (type(key) == 'string' and
                         not reserved_keys[key] and match(key, '^[%a_][%d%a_]*$') and
                         key or '[' .. visit(key) .. ']') ..
                    equals .. visit(ref[key]))
            end
            keys = nil -- Free the temp table no longer needed.
            -- Finally dump the metatable (with pseudo-key '[]'), if there's one.
            if include_mt then
                ref = getmetatable(ref)
                if ref then
                    insert(result, '[]' .. equals .. visit(ref))
                end
            end
            -- Pack the result string.
            -- TODO: improve pretty-printing with newlines/indentation
            return open .. concat(result, comma) .. close
        elseif typ ~= 'nil' then -- other reference types (function, userdata, etc.)
            local id = ref_ids[ref]
            if not id then
                id = (ref_counts[typ] or 0) + 1; ref_ids[ref], ref_counts[ref] = id, id
            end
            return ':' .. typ .. '#' .. id .. ':'
        else
            return 'nil'
        end
    end
    return visit(val)
end